@@ -26,9 +26,9 @@
 #include <io.h>
 #endif
 
-#include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 
 #if !defined(UNIX) && !defined(SASC)
 #include <share.h>
@@ -38,6 +38,11 @@
 #include <dos.h>
 #endif
 
+#ifdef __BEOS__
+#include <be/kernel/fs_attr.h>
+#include <be/support/TypeConstants.h>
+#endif
+
 #define MSGAPI_HANDLERS
 
 #include "prog.h"
@@ -144,6 +149,36 @@ ErrOpen:
     return NULL;
 }
 
+int SdmDeleteBase(char *name)
+{
+    FFIND *ff;
+    char *temp;
+
+    temp = malloc(strlen(name)+6);
+    sprintf(temp, "%s*.msg", name);
+
+    ff = FFindOpen(temp, 0);
+
+    free(temp);
+
+    if (ff != 0)
+    {
+      do
+	  {
+	    temp = malloc(strlen(name) + strlen(ff->ff_name) + 1);
+	    sprintf(temp, "%s%s", name, ff->ff_name);
+	    unlink(temp);
+	    free(temp);
+	  }
+      while (FFindNext(ff) == 0);
+
+      FFindClose(ff);
+    }
+    rmdir(name);
+
+    return 1; /* rmdir error is ok */
+}
+
 static sword EXPENTRY SdmCloseArea(MSG * mh)
 {
     XMSG msg;
@@ -355,11 +390,8 @@ static MSGH *EXPENTRY SdmOpenMsg(MSG * mh, word mode, dword msgnum)
     }
 
     sprintf((char *) msgname, (char *) sd_msg, Mhd->base, (int)msgnum);
-#ifdef UNIX
-    handle = sopen((char *) msgname, filemode | O_BINARY, SH_DENYNONE, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
-#else
-    handle = sopen((char *) msgname, filemode | O_BINARY, SH_DENYNONE, S_IREAD | S_IWRITE);
-#endif
+    handle = sopen((char *) msgname, filemode | O_BINARY, SH_DENYNONE,
+                   FILEMODE(mh->isecho));
     if (handle == -1)
     {
         if (filemode & O_CREAT)
@@ -398,7 +430,7 @@ static MSGH *EXPENTRY SdmOpenMsg(MSG * mh, word mode, dword msgnum)
                 pfree(msgh);
                 close(handle);
                 msgapierr = MERR_NOMEM;
-                return NULL;
+		return NULL;
             }
         }
 
@@ -409,7 +441,8 @@ static MSGH *EXPENTRY SdmOpenMsg(MSG * mh, word mode, dword msgnum)
 
         if (!owrite)
         {
-            Mhd->msgnum[(size_t) (mh->num_msg++)] = (word) msgnum;
+	    Mhd->msgnum[(size_t) (mh->num_msg)] = (word) msgnum;
+	    mh->num_msg++;
         }
         else
         {
@@ -593,6 +626,7 @@ static dword EXPENTRY SdmReadMsg(MSGH * msgh, XMSG * msg, dword offset, dword by
         }
 
         got = (word) farread(msgh->fd, text, (unsigned int)bytes);
+        text[(unsigned int)got] = '\0';
 
         /*
          *  Update counter only if we got some text, and only if we're
@@ -645,6 +679,7 @@ static dword EXPENTRY SdmReadMsg(MSGH * msgh, XMSG * msg, dword offset, dword by
 
         slen = strlen((char *) msgh->ctrl) + 1;
         memmove(ctxt, msgh->ctrl, min(slen, (size_t) clen));
+        ctxt[min(slen, (size_t) clen)] = '\0';
     }
 
     if (fake_msgbuf)
@@ -692,6 +727,24 @@ static sword EXPENTRY SdmWriteMsg(MSGH * msgh, word append, XMSG * msg, byte * t
             statfd = msgh->fd;
             msgh->zplen = (word) WriteZPInfo(msg, WriteToFd, ctxt);
         }
+        
+        /* Use Attributes und BeOS */
+#ifdef __BEOS__
+        {
+        struct tm tmdate;
+        time_t ttime;
+        
+        fs_write_attr(msgh->fd, "BEOS:TYPE", B_MIME_TYPE, 0l, "message/fmsg", 13);
+        fs_write_attr(msgh->fd, "XMSG:FROM", B_STRING_TYPE, 0l, msg->from, strlen(msg->from));
+        fs_write_attr(msgh->fd, "XMSG:TO"  , B_STRING_TYPE, 0l, msg->to, strlen(msg->to));
+        fs_write_attr(msgh->fd, "XMSG:SUBJ", B_STRING_TYPE, 0l, msg->subj, strlen(msg->subj));
+        ttime = mktime(DosDate_to_TmDate((union stamp_combo *)&msg->date_written, &tmdate));
+        fs_write_attr(msgh->fd, "XMSG:DATE", B_TIME_TYPE, 0l, &ttime, 4l);
+        ttime = mktime(DosDate_to_TmDate((union stamp_combo *)&msg->date_arrived, &tmdate));
+        fs_write_attr(msgh->fd, "XMSG:DTAR", B_TIME_TYPE, 0l, &ttime, 4l);
+        /* ... and so on ... not fully implemented ! (Yet) */
+        }
+#endif
     }
     else if (!append || ctxt)
     {
@@ -739,6 +792,12 @@ static sword EXPENTRY SdmWriteMsg(MSGH * msgh, word append, XMSG * msg, byte * t
             msgapierr = MERR_NODS;
             return -1;
         }
+        if (text[textlen])
+            if (farwrite(msgh->fd, "", 1) != 1)
+            {
+                msgapierr = MERR_NODS;
+                return -1;
+            }
     }
 
     msgapierr = MERR_NONE;
@@ -1041,7 +1100,7 @@ static sword near _SdmRescanArea(MSG * mh)
 
     Mhd->msgnum_len = SDM_BLOCK;
 
-    temp = malloc(strlen(Mhd->base)+6);
+    temp = malloc(strlen((char *)Mhd->base)+6);
     sprintf((char *) temp, "%s*.msg", Mhd->base);
 
     ff = FFindOpen((char *) temp, 0);
@@ -1089,24 +1148,8 @@ static sword near _SdmRescanArea(MSG * mh)
             }
 
 #ifdef OS2
-#ifdef __TURBOC__
             if ((mn % 128) == 127)
-            {
-                sleep(1);
-            }
-#else
-            {
-#ifdef __IBMC__
-                extern void _System DosSleep(dword);
-#else
-                extern void far pascal DosSleep(dword);
-#endif
-                if ((mn % 128) == 127)
-                {
-                    DosSleep(1L);
-                }
-            }
-#endif
+                tdelay(1L); /* give up cpu */
 #endif
         }
         while (FFindNext(ff) == 0);
@@ -1284,3 +1327,38 @@ static void near Get_Binary_Date(struct _stamp *todate, struct _stamp *fromdate,
         *todate = *fromdate;
     }
 }
+
+static dword EXPENTRY SdmGetHash(HAREA mh, dword msgnum)
+{
+  XMSG xmsg;
+  HMSG msgh;
+  dword rc = 0l;
+
+  if ((msgh=SdmOpenMsg(mh, MOPEN_READ, msgnum))==NULL)
+    return (dword) 0l;
+
+  if (SdmReadMsg(msgh, &xmsg, 0L, 0L, NULL, 0L, NULL)!=(dword)-1)
+  {
+    rc = SquishHash(xmsg.to) | (xmsg.attr & MSGREAD) ? 0x80000000l : 0;
+  }
+
+  SdmCloseMsg(msgh);
+
+  msgapierr=MERR_NONE;
+  return rc;
+}
+
+static UMSGID EXPENTRY SdmGetNextUid(HAREA ha)
+{
+  if (InvalidMh(ha))
+    return 0L;
+
+  if (!ha->locked)
+  {
+    msgapierr=MERR_NOLOCK;
+    return 0L;
+  }
+
+  msgapierr=MERR_NONE;
+  return ha->high_msg+1;
+}
