@@ -48,9 +48,9 @@
 #include <unistd.h>
 #endif
 
-#include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 
 #define MSGAPI_HANDLERS
 #include "prog.h"
@@ -69,10 +69,10 @@
  */
 
 #define get_dword(ptr)            \
-   ((dword)((ptr)[0]) |           \
-    (((dword)((ptr)[1])) << 8)  | \
-    (((dword)((ptr)[2])) << 16) | \
-    (((dword)((ptr)[3])) << 24))  \
+   ((dword)((unsigned char)(ptr)[0]) |           \
+    (((dword)((unsigned char)(ptr)[1])) << 8)  | \
+    (((dword)((unsigned char)(ptr)[2])) << 16) | \
+    (((dword)((unsigned char)(ptr)[3])) << 24))  \
 
 /*
  *  get_word
@@ -83,9 +83,8 @@
  */
 
 #define get_word(ptr)         \
-    ((word)(ptr)[0] |         \
-     (((word)(ptr)[1]) << 8 ))
-
+    ((word)((unsigned char)(ptr)[0]) |         \
+     (((word)((unsigned char)(ptr)[1])) << 8 ))
 
 /*
  *  put_dword
@@ -94,7 +93,7 @@
  *  system architecture.
  */
 
-static void put_dword(byte *ptr, dword value)
+void put_dword(byte *ptr, dword value)
 {
     ptr[0] = (value & 0xFF);
     ptr[1] = (value >> 8) & 0xFF;
@@ -109,7 +108,7 @@ static void put_dword(byte *ptr, dword value)
  *  system architecture.
  */
 
-static void put_word(byte *ptr, word value)
+void put_word(byte *ptr, word value)
 {
     ptr[0] = (value & 0xFF);
     ptr[1] = (value >> 8) & 0xFF;
@@ -189,6 +188,10 @@ int read_xmsg(sword handle, XMSG *pxmsg)
         pbuf += 4;
     }
 
+                                /* 4 bytes "umsgid" */
+    pxmsg->umsgid = get_dword(pbuf);
+    pbuf += 4;
+
                                 /* 20 times FTSC date stamp */
     memmove(pxmsg->__ftsc_date, pbuf, 20);
     pbuf += 20;
@@ -276,6 +279,10 @@ int write_xmsg(sword handle, XMSG *pxmsg)
         put_dword(pbuf, pxmsg->replies[i]);
         pbuf += 4;
     }
+                                /* 4 bytes "umsgid" */
+    put_dword(pbuf, pxmsg->umsgid);
+    pbuf += 4;
+
 
                                 /* 20 times FTSC date stamp */
     memmove(pbuf, pxmsg->__ftsc_date, 20);
@@ -510,6 +517,11 @@ int write_sqidx(sword handle, SQIDX *psqidx, dword n)
         }
     }
 
+    if (accel_buffer != NULL)
+    {
+        free(accel_buffer);
+    }
+
     return 1;
 }
 
@@ -865,7 +877,7 @@ int read_hdr(sword handle, JAMHDR *Hdr)
    } /* endif */
 
    /* 04 bytes Signature */
-   memmove(Hdr->Signature, pbuf, (size_t)4); 
+   memmove(Hdr->Signature, pbuf, (size_t)4);
    pbuf += 4;
 
    /* 02 bytes Revision */
@@ -949,24 +961,34 @@ int read_hdr(sword handle, JAMHDR *Hdr)
    return 1;
 }
 
-int read_subfield(sword handle, JAMSUBFIELDptr *subfield, dword SubfieldLen)
+int read_subfield(sword handle, JAMSUBFIELD2ptr *subfield, dword *SubfieldLen)
 {
    char *buf, *pbuf;
-   JAMSUBFIELDptr subfieldNext;
+   JAMSUBFIELD2ptr subfieldNext, subfieldNew;
    dword datlen;
 
-   buf = (char*)palloc(SubfieldLen);
+   buf = (char*)palloc(*SubfieldLen);
 
    pbuf = buf;
 
-   if (farread(handle, (byte far *)buf, SubfieldLen) != SubfieldLen) {
+   if (farread(handle, (byte far *)buf, *SubfieldLen) != *SubfieldLen) {
       pfree(buf);
       return 0;
    } /* endif */
 
-   subfieldNext = *subfield;
+   subfieldNext = *subfield = 0;
+
+   while ((pbuf - buf + 8) < *SubfieldLen)
+	 {
+      subfieldNew = (JAMSUBFIELD2ptr) palloc(sizeof(JAMSUBFIELD2));
+			subfieldNew->next = 0;
+			if (subfieldNext)
+			  subfieldNext->next = subfieldNew;
+			subfieldNext = subfieldNew;
+
+			if (!*subfield)
+			  *subfield = subfieldNext;
 
-   while ((pbuf - buf) < SubfieldLen) {
       /* 02 bytes LoID */
       subfieldNext->LoID = get_word(pbuf);
       pbuf += 2;
@@ -976,17 +998,28 @@ int read_subfield(sword handle, JAMSUBFIELDptr *subfield, dword SubfieldLen)
       pbuf += 2;
 
       /* 04 bytes DatLen */
-      datlen = subfieldNext->DatLen = get_dword(pbuf);
+      subfieldNext->DatLen = 0;
+      datlen = get_dword(pbuf);
       pbuf += 4;
 
+      if (pbuf - buf + datlen > *SubfieldLen)
+          break;
       /* DatLen bytes Buffer */
-      memmove(subfieldNext->Buffer, pbuf, datlen);
+      if ((long)datlen >= 0 && datlen < *SubfieldLen)
+      { subfieldNext->DatLen = datlen;
+			  subfieldNext->Buffer = malloc(datlen+1);
+        memmove(subfieldNext->Buffer, pbuf, datlen);
+      }
+      else
+        break;
       pbuf += datlen;
 
-      subfieldNext = (JAMSUBFIELDptr)((char *)subfieldNext + sizeof(JAMBINSUBFIELD) + datlen);
+/*      subfieldNext = (JAMSUBFIELDptr)((char *)subfieldNext + sizeof(JAMBINSUBFIELD) + datlen);
+*/
 
    } /* endwhile */
 
+   *SubfieldLen = pbuf - buf;
    pfree(buf);
 
    return 1;
@@ -1003,7 +1036,7 @@ int read_allidx(JAMBASEptr jmb)
    len = tell(jmb->IdxHandle);
    lseek(jmb->IdxHandle, 0, SEEK_SET);
 
-   buf = (char*)palloc(len);
+   buf = (unsigned char*)palloc(len);
    pbuf = buf;
 
    if (farread(jmb->IdxHandle, (byte far *)buf, len) != len) {
@@ -1014,7 +1047,7 @@ int read_allidx(JAMBASEptr jmb)
 
    for (i = 0; (pbuf - buf) < len;) {
       offset = get_dword(pbuf+4);
-      if (offset != 0xFFFFFFFF) {
+      if (offset != 0xFFFFFFFFUL) {
          lseek(jmb->HdrHandle, offset, SEEK_SET);
          if (farread(jmb->HdrHandle, (byte far *)hbuf, HDR_SIZE) == HDR_SIZE) {
             attr = get_dword(phbuf+52);
@@ -1102,7 +1135,7 @@ int write_hdr(sword handle, JAMHDR *Hdr)
    byte buf[HDR_SIZE], *pbuf = buf;
 
    /* 04 bytes Signature */
-   memmove(pbuf, Hdr->Signature, (size_t)4); 
+   memmove(pbuf, Hdr->Signature, (size_t)4);
    pbuf += 4;
 
    /* 02 bytes Revision */
@@ -1186,41 +1219,39 @@ int write_hdr(sword handle, JAMHDR *Hdr)
    return (farwrite(handle, (byte far *)buf, HDR_SIZE) == HDR_SIZE);
 }
 
-int write_subfield(sword handle, JAMSUBFIELDptr *subfield, dword SubfieldLen)
+int write_subfield(sword handle, JAMSUBFIELD2ptr *subfield, dword SubfieldLen)
 {
-   char *buf, *pbuf;
+   unsigned char *buf, *pbuf;
    dword datlen;
-   JAMSUBFIELDptr subfieldNext;
+   JAMSUBFIELD2ptr subfieldNext;
+	 int rc;
 
-   buf = (char*)palloc(SubfieldLen);
+
+   buf = (unsigned char*)palloc(SubfieldLen);
    pbuf = buf;
    subfieldNext = *subfield;
-
-   while ((pbuf - buf) < SubfieldLen) {
-
+   while (subfieldNext) {
       /* 02 bytes LoID */
       put_word(pbuf, subfieldNext->LoID);
       pbuf += 2;
-
       /* 02 bytes HiID */
       put_word(pbuf, subfieldNext->HiID);
       pbuf += 2;
-
       /* 04 bytes DatLen */
       put_dword(pbuf, subfieldNext->DatLen);
+
       datlen = subfieldNext->DatLen;
       pbuf += 4;
-
       /* DatLen bytes Buffer */
       memmove(pbuf, subfieldNext->Buffer, datlen);
       pbuf += datlen;
-
-      subfieldNext = (JAMSUBFIELDptr)((char *)subfieldNext + sizeof(JAMBINSUBFIELD) + datlen);
+/*      subfieldNext = (JAMSUBFIELDptr)((char *)subfieldNext + sizeof(JAMBINSUBFIELD) + datlen);*/
+      subfieldNext = subfieldNext->next;
 
    } /* endwhile */
+   rc =(farwrite(handle, (byte far *)buf, SubfieldLen) == SubfieldLen);
 
-   return (farwrite(handle, (byte far *)buf, SubfieldLen) == SubfieldLen);
+   pfree(buf);
+
+   return rc;
 }
-
-
-
